unit ICQPkt;
(***********************************************************)
(*   TICQPkt Object, v. 01 (initial release)
(*   Object used to create login/logoff packets and 
(*   also parses login and header parts of a packet.
(*
(*   Written by Coban(alex@ritlabs.com), special for using
(*   in ICQ Password Whacker app.
(**********)
interface
uses
  SysUtils,
  ICQCrypt;

const
  ICQ_TCP_VER = $0600;
  ICQ_BUILD_DATE = $3B8CED1A;
  ICQ_X1 = $d5000000;
  ICQ_X2 = $0800d500;
  ICQ_X3 = $50000000;
  ICQ_X4 = $03000000;
  ICQ_QUITMSG = 'B_USER_DISCONNECTED';
  HexDigts: String[16] = '0123456789ABCDEF';

type
  TSRV_HEADER = record
    srv_ver: Word;
    srv_sess_id: LongWord;
    srv_command: Byte;
    srv_seq1: Byte;
    srv_seq2: Byte;
    srv_uin: LongWord;
    srv_checkcode: LongWord;
  end;

  TSRV_LOGIN_REPLY = record
    srv_x1: LongWord;
    srv_x2: Word;
    srv_x3: Word;
    srv_x4: Word;
    srv_x5: Word;
    srv_ip: LongWord;
    srv_x6: LongWord;
  end;

  TICQPkt = class
  private
    pkt: array [0..1024] of Byte;
    pkt_len: Word;
    procedure StrAppend(var pkt_arr: array of Byte; StartPos: Word; EndPos: Word; Value: String);
    procedure HexAppend(var pkt_arr: array of Byte; StartPos: Word; EndPos: Word; Value: String);
    function Reverse(N: LongWord): LongWord;
//    function HexToInt(S: String): LongWord;
  public
    constructor Create;
    destructor Destroy; override;
    procedure BuildHeader(ICQ_UIN, ICQ_SESSID: LongWord; ICQ_COMMAND, ICQ_SEQ1, ICQ_SEQ2: Word);
    procedure BuildLoginPkt(LPort: LongWord; Password: String; Pass_len: Byte; LIP: LongWord; Status: LongWord);
    procedure BuildDisconnectPkt;
    procedure ResetPacket;
    function ParseHeader(var buf: array of Byte): TSRV_HEADER;
    function ParseLoginReply(var buf: array of Byte): TSRV_LOGIN_REPLY;
    procedure GetPktStr(var buf: array of Char; var buf_len: Integer);
    procedure GetPktByte(var buf: array of Byte; var buf_len: Integer);
  end;

implementation
  constructor TICQPkt.Create;
  begin
    pkt_len := 0;
    FillChar(pkt, SizeOf(pkt), 0);
  end;

  destructor TICQPkt.Destroy;
  begin
  end;

  procedure TICQPkt.BuildHeader(ICQ_UIN, ICQ_SESSID: LongWord; ICQ_COMMAND, ICQ_SEQ1, ICQ_SEQ2: Word);
  begin
    HexAppend(pkt, 0, 2, '0500');                              //version
    HexAppend(pkt, 2, 6, '00000000');                          //zeros
    HexAppend(pkt, 6, 10, IntToHex(Reverse(ICQ_UIN), 8));      //uin
    HexAppend(pkt, 10, 14, IntToHex(Reverse(ICQ_SESSID), 8));  //session id
    HexAppend(pkt, 14, 16, IntToHex(ICQ_COMMAND, 4));          //command
    HexAppend(pkt, 16, 18, IntToHex(ICQ_SEQ1, 4));             //seq1
    HexAppend(pkt, 18, 20, IntToHex(ICQ_SEQ2, 4));             //seq2
    HexAppend(pkt, 20, 24, '00000000');                        //zero filled place for checkcode
  end;

  procedure TICQPkt.BuildLoginPkt(LPort: LongWord; Password: String; Pass_len: Byte; LIP: LongWord; Status: LongWord);
  begin
    HexAppend(pkt, 24, 28, IntToHex($3B8CED1B, 4));                                      //current time (not used in processing)
    HexAppend(pkt, 28, 32, IntToHex(Reverse(LPort), 8));                                 //local port
    HexAppend(pkt, 32, 34, IntToHex(Reverse(Pass_len + 1), 8));                          //password length (length + 1 for #0 char)
    StrAppend(pkt, 33, 34 + Pass_len, Password + #0);                                    //null terminated password
    HexAppend(pkt, 34 + Pass_len + 1, 35 + Pass_len + 4, IntToHex(ICQ_X1, 4));           //x1
    HexAppend(pkt, 34 + Pass_len + 5, 34 + Pass_len + 9, IntToHex(Reverse(LIP), 8));     //ip
    HexAppend(pkt, 34 + Pass_len + 9, 34 + Pass_len + 10, IntToHex($04, 2));             //flag
    HexAppend(pkt, 34 + Pass_len + 10, 34 + Pass_len + 14, IntToHex(Status, 8));         //status
    HexAppend(pkt, 34 + Pass_len + 14, 34 + Pass_len + 16, IntToHex(ICQ_TCP_VER, 4));    //icq version
    HexAppend(pkt, 34 + Pass_len + 16, 34 + Pass_len + 22, '000000000000');              //just zeros
    HexAppend(pkt, 34 + Pass_len + 22, 34 + Pass_len + 26, IntToHex(ICQ_X2, 8));         //x1
    HexAppend(pkt, 34 + Pass_len + 26, 34 + Pass_len + 30, IntToHex(ICQ_X3, 8));         //x2
    HexAppend(pkt, 34 + Pass_len + 30, 34 + Pass_len + 34, IntToHex(ICQ_X4, 8));         //x3
    HexAppend(pkt, 34 + Pass_len + 34, 34 + Pass_len + 38, IntToHex(ICQ_BUILD_DATE, 8)); //icq build date (it seems mirabilis do not use this property)}
  end;

  procedure TICQPkt.BuildDisconnectPkt;
  begin
    HexAppend(pkt, 24, 25, IntToHex(Reverse(Length(ICQ_QUITMSG) + 1), 4));               //quit message length
    StrAppend(pkt, 25, 25 + Length(ICQ_QUITMSG) + 2, ICQ_QUITMSG + #0);                  //quit message text
    HexAppend(pkt, 25 + Length(ICQ_QUITMSG) + 2, 25 + Length(ICQ_QUITMSG) + 4, '0500');  //0500
  end;

  function TICQPkt.ParseHeader(var buf: array of Byte): TSRV_HEADER;
  var
    sh: TSRV_HEADER;
  begin
    FillChar(sh, SizeOf(sh), 0);
    sh.srv_ver := Word(buf[1] shl 8 + buf[0]);
    sh.srv_sess_id := LongWord(buf[6] shl 24 + buf[5] shl 16 + buf[4] shl 8 + buf[3]);
    sh.srv_command := Word(buf[8] shl 8 + buf[7]);
    sh.srv_seq1 := Word(buf[10] shl 8 + buf[9]);
    sh.srv_seq2 := Word(buf[11] shl 8 + buf[10]);
    sh.srv_uin := LongWord(buf[15] shl 24 + buf[14] shl 16 + buf[13] shl 8 + buf[12]);
    sh.srv_checkcode := LongWord(buf[19] shl 24 + buf[18] shl 16 + buf[17] shl 8 + buf[16]);
    Result := sh;
  end;

  function TICQPkt.ParseLoginReply(var buf: array of Byte): TSRV_LOGIN_REPLY;
  var
    sr: TSRV_LOGIN_REPLY;
  begin
    FillChar(sr, SizeOf(sr), 0);
    sr.srv_x1 := LongWord(buf[3] shl 24 + buf[2] shl 16 + buf[1] shl 8 + buf[0]);
    sr.srv_x2 := Word(buf[5] shl 8 + buf[4]);
    sr.srv_x3 := Word(buf[7] shl 8 + buf[6]);
    sr.srv_x4 := Word(buf[9] shl 8 + buf[8]);
    sr.srv_x5 := Word(buf[11] shl 8 + buf[10]);
    sr.srv_ip := LongWord(buf[15] shl 24 + buf[14] shl 16 + buf[13] shl 8 + buf[12]);
    sr.srv_x6 := LongWord(buf[19] shl 24 + buf[18] shl 16 + buf[17] shl 8 + buf[16]);
    Result := sr;
  end;

  procedure TICQPkt.ResetPacket;
  begin
    TICQPkt.Create;
  end;

  procedure TICQPkt.GetPktStr(var buf: array of Char; var buf_len: Integer);
  var
    i: Integer;
  begin
    FillChar(buf, SizeOf(buf), 0);
    for i := Low(pkt) to pkt_len do
      if SizeOf(buf) < i then
        Continue
      else
        buf[i] := Chr(pkt[i]);
    buf_len := pkt_len;
  end;

  procedure TICQPkt.GetPktByte(var buf: array of Byte; var buf_len: Integer);
  var
    i: Integer;
  begin
    FillChar(buf, SizeOf(buf), 0);
    for i := Low(pkt) to pkt_len do
      if SizeOf(buf) < i then
        Continue
      else
        buf[i] := pkt[i];
    buf_len := pkt_len;
  end;

  function TICQPkt.Reverse(N: LongWord): LongWord;
  var
    Res: LongWord;
  begin
    Res := 0;
    Res := Res + (N and $FF000000) shr 24;
    Res := Res + (N and $00FF0000) shr 8;
    Res := Res + (N and $0000FF00) shl 8;
    Res := Res + (N and $000000FF) shl 24;
    Result := Res;
  end;

  procedure TICQPkt.HexAppend(var pkt_arr: array of Byte; StartPos: Word; EndPos: Word; Value: String);
  var
    i: Integer;
    j: Integer;
    num: Byte;
    num2: Byte;
  begin
    j := StartPos;
    while (EndPos - StartPos) * 2 > Length(Value) do
      Value := Value + '0';
    for i := 0 to (EndPos - StartPos) * 2 do
    begin
      if i mod 2 = 0 then
      begin
        //check for overflow
        if i + 1 > (EndPos - StartPos) * 2 then
          Continue;
        if Pos(Value[i + 1], HexDigts) > 0 then
          num := Pos(Value[i + 1], HexDigts) - 1
        else
          num := 0;
        //check for overflow
        if i + 2 > (EndPos - StartPos) * 2 then
          Continue;
        if Pos(Value[i + 2], HexDigts) > 0 then
          num2 := Pos(Value[i + 2], HexDigts) - 1
        else
          num2 := 0;
        pkt_arr[j] := num * 16 + num2;
        Inc(j);
        Inc(pkt_len);
      end;
    end;
  end;

  procedure TICQPkt.StrAppend(var pkt_arr: array of Byte; StartPos: Word; EndPos: Word; Value: String);
  var
    i: Integer;
  begin
    while (EndPos - StartPos) > Length(Value) do
      Value := Value + '0';
    for i := 1 to (EndPos - StartPos) do
    begin
      pkt_arr[StartPos + i] := Ord(Value[i]);
      Inc(pkt_len);
    end;
  end;

{  function TICQPkt.HexToInt(S: String): LongWord;
   var
       i: Byte;
   begin
     Result := 0;
     for i := 1 to Length(S) do
       Inc(Result, (1 shl (4 * (Length(S) - i))) * Pos(S, HexDigts));
  end;}
end.